/* 

Copyright (c) 1999  Ulric Eriksson <ulric@siag.nu>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <magick/magick.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <limits.h>

#ifndef USHRT_MAX
#define USHRT_MAX 65535
#endif

static int rgb_shift = 11;

/* known colours are cached in a sorted list to avoid
   making multiple calls to the X server */
/* Beware though: alloc_color sorts on rgb, query_color does not sort */
typedef struct xmagick_color {
	XColor color;
	unsigned long rgb;	/* (up to) 24-bit colour */
} xmagick_color;

static xmagick_color *known_colors = NULL;

static int ncolors, maxcolors, lastcolor;

/* forget all cached colours */
static void free_colors(void)
{
	free(known_colors);
	known_colors = NULL;
	ncolors = 0;
	lastcolor = -1;
}

/* call XAllocColor through the cache */
/* Does not fail. Returns closest match at worst. */
/* Algorithm:
   1. If same rgb value as last time, return same pixel value.
   2. If exact match found, return it.
   3. Call XAllocColor. If successful, return it.
   4. Return "anything".

   If N is the number of picture elements (width*height), M is the number
   of picture elements that are the same as the one before, C is the number
   of unique colours, we get O((N-M)*log(C) + C*C + M*1). Those come from
   number of table lookups + number of colour insertions + "same" colours.
*/
static void alloc_color(Display *display, Colormap colormap, XColor *color)
{
	int i, j, lower, upper;
	int cr, cg, cb, d;
	unsigned long rgb;

	cr = color->red >> rgb_shift;
	cg = color->green >> rgb_shift;
	cb = color->blue >> rgb_shift;
	rgb = (cr << 16) + (cg << 8) + cb;

	/* take advantage of long stretches of identical pixels */
	if (lastcolor != -1 && known_colors[lastcolor].rgb == rgb) {
		color->pixel = known_colors[lastcolor].color.pixel;
		return;
	}
	/* binary search through known colours */
	lower = 0;
	upper = ncolors-1;
	while (lower <= upper) {
		i = (lower+upper)/2;
		d = rgb-known_colors[i].rgb;
		if (d == 0) {
			lastcolor = i;
			color->pixel = known_colors[i].color.pixel;
			return;
		} else if (d > 0) {
			lower = i+1;
		} else {
			upper = i-1;
		}
	}
	/* if we are here, then lower == upper+1 and no match found */
	if (ncolors >= maxcolors) {	/* too many colours */
		/* no reason to change lastcolor */
		color->pixel = 0;
		return;
	}
	if (XAllocColor(display, colormap, color) == 0) {
		int dr, dg, db, bd = INT_MAX, best = 0;
		/* linear search, yuck */
		for (j = 0; j < ncolors; j++) {
			XColor c = known_colors[j].color;
			dr = cr - (c.red >> rgb_shift);
			dg = cg - (c.green >> rgb_shift);
			db = cb - (c.blue >> rgb_shift);
			d = dr*dr+dg*dg+db*db;
			if (d < bd) {
				bd = d;
				best = j;
			}
		}
		color->pixel = known_colors[best].color.pixel;
	}
	/* insert new colour */
	if (known_colors[upper].rgb > rgb) i = upper;
	else i = lower;
	if (i < 0) i = 0;		/* before everything */
	if (i > ncolors) i = ncolors;	/* after everything */
	for (j = ncolors; j > i; j--) {
		known_colors[j] = known_colors[j-1];
	}
	known_colors[i].color = *color;
	known_colors[i].rgb = rgb;
	known_colors[i].color.pixel = color->pixel;
	lastcolor = i;
	ncolors++;
}

/* call XQueryColor through the cache */
static void query_color(Display *display, Colormap colormap, XColor *color)
{
	int i;

	for (i = 0; i < ncolors; i++) {
		if (known_colors[i].color.pixel == color->pixel) {
			color->red = known_colors[i].color.red;
			color->green = known_colors[i].color.green;
			color->blue = known_colors[i].color.blue;
			return;
		}
	}
	if (i >= maxcolors) {
		color->red = 0;
		color->green = 0;
		color->blue = 0;
		return;
	}
	XQueryColor(display, colormap, color);
	known_colors[i].color = *color;
	ncolors++;
}

/* Convert Image to XImage */
XImage *XMagickImageToXImage(Display *display, Image *image)
{
	XColor color;
	unsigned long pixel;
	int i;
	float *red, *green, *blue, *opacity;
	int x, y;
	XImage *ximage = NULL;
	int screen = DefaultScreen(display);
	Visual *visual = DefaultVisual(display, screen);
	int format = ZPixmap;
	int offset = 0;
	char *data = NULL;
	Colormap colormap = DefaultColormap(display, screen);
	int bitmap_pad = XBitmapPad(display);
	int bytes_per_line = 0;
	unsigned int width = image->columns;
	unsigned int height = image->rows;
	unsigned int depth = DefaultDepth(display, screen);

	maxcolors = GetNumberColors(image, NULL);
	free_colors();
	known_colors = malloc(maxcolors*sizeof *known_colors);
	red = malloc((width*height)*(sizeof *red));
	green = malloc((width*height)*(sizeof *green));
	blue = malloc((width*height)*(sizeof *blue));
	opacity = malloc((width*height)*(sizeof *opacity));

	if (red && green && blue && opacity) {
		GetPixels(image, red, green, blue, opacity);
		ximage = XCreateImage(display, visual, depth,
				format, offset, data,
				width, height, bitmap_pad, bytes_per_line);
		ximage->data = malloc(ximage->bytes_per_line
					*ximage->height*ximage->depth);
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				i = x+y*width;
				color.red = USHRT_MAX*red[i];
				color.green = USHRT_MAX*green[i];
				color.blue = USHRT_MAX*blue[i];
				alloc_color(display, colormap, &color);
				pixel = color.pixel;
				XPutPixel(ximage, x, y, pixel);
			}
		}
	}

	free(red);
	free(green);
	free(blue);
	free(opacity);

	free_colors();
	return ximage;
}

/* Convert XImage to Image */
Image *XMagickXImageToImage(Display *display, XImage *ximage)
{
	XColor color;
	unsigned long pixel;
	int i;
	float *red, *green, *blue, *opacity;
	Image *image = NULL;
	int x, y;
	int screen = DefaultScreen(display);
	Colormap colormap = DefaultColormap(display, screen);
	unsigned int width = ximage->width;
	unsigned int height = ximage->height;

	free_colors();
	maxcolors = 65555;	/* arbitrary */
	known_colors = malloc(maxcolors*sizeof *known_colors);
	red = malloc((width*height)*(sizeof *red));
	green = malloc((width*height)*(sizeof *green));
	blue = malloc((width*height)*(sizeof *blue));
	opacity = malloc((width*height)*(sizeof *opacity));

	if (red && green && blue && opacity) {
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				i = x+y*width;
				pixel = XGetPixel(ximage, x, y);
				color.pixel = pixel;
				query_color(display, colormap, &color);
				red[i] = (float)color.red/USHRT_MAX;
				green[i] = (float)color.green/USHRT_MAX;
				blue[i] = (float)color.blue/USHRT_MAX;
				opacity[i] = 1.0;
			}
		}
		image = CreateImage(width, height, red, green, blue, opacity);
	}

	free(red);
	free(green);
	free(blue);
	free(opacity);

	free_colors();
	return image;
}

/* Set the number of bits to use when calculating the rgb value for
   alloc_color. Valid values are 1 to 8. Returns the new value.
*/
int XMagickRGBBits(int bits)
{
	if (bits < 1) bits = 1;
	else if (bits > 8) bits = 8;
	rgb_shift = 16-bits;
	return bits;
}

